/*
 *
 * Diaconu Cristian
 * 333 CA
 * Proiect PM 2017
 *
*/

#define F_CPU 16000000
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BUFF_SIZE_X      (LCD_X_RES)
#define BUFF_SIZE_Y      (LCD_Y_RES / 8)

#define	WIDTH	14
#define	HEIGHT	6

#define	LEFT	1
#define	RIGHT	2
#define	DOWN	3
#define	UP		4
#define NONE	5

#define FOOD_BONUS_DELAY    4
#define FOOD_POISON_DELAY   3
#define SNAKE_START_LENGTH  3
#define FOOD_VALUE          5
#define BONUS_VALUE        15
#define POISON_VALUE      -10
#define LEVEL_UPGRADE      10

#define NUM_OBSTACLES       5


typedef struct coord {
	int x;
	int y;
} coord;


/* Global parameters */
static coord snake[WIDTH * HEIGHT];
static coord food;
static coord obstacles[NUM_OBSTACLES];
static unsigned int snake_size;  //former size
static unsigned int food_count;	//former foodNum
static int food_value;   //former foodValue
static int score;
static int direction;
static int level;
static int single_level;


/* Init buttons */
void enable_buttons(void)
{
	DDRB = 0xff;
	PORTB = 0x00;

	/* Set pins as input */
	DDRB &= ~(1 << PB0);
	DDRB &= ~(1 << PB1);
	DDRB &= ~(1 << PB2);
	DDRB &= ~(1 << PB3);

	/* Enable pull-up for pins. */
	PORTB |= (1 << PB0);
	PORTB |= (1 << PB1);
	PORTB |= (1 << PB2);
	PORTB |= (1 << PB3);
}


/* Randomly set bonuses or food or poison on board */
void set_food(void)
{
	unsigned int i;
	
	food_count++;
	
	/* level 3 only - randomness when generating bonuses */
	if (level == 3) {
		i = rand() % 3;
		if (i == 0)
			food_value = FOOD_VALUE;
		else if (i == 1)
			food_value = BONUS_VALUE;
		else
			food_value = POISON_VALUE;
	}
	/* bonus */
	else if (food_count % FOOD_BONUS_DELAY == 0)
		food_value = BONUS_VALUE;
	/* regular food */
	else
		food_value = FOOD_VALUE;
	
	food.x = rand() % (WIDTH - 1)  + 1;
	food.y = rand() % (HEIGHT - 1) + 1;
	
	/* if food or bonus is generated on snake surface */
	/* we go backwards and generate another one */
	for (i = 0; i < snake_size; i++) {
		if (snake[i].x == food.x && snake[i].y == food.y) {
			food_count--;
			set_food();
		}
	}
	
	/* set food for levels 2 and 3*/
	if (level > 1) {
	
		for (i = 0; i < NUM_OBSTACLES; i++) {
			if ((food.x == obstacles[i].x) ||
				(food.y == obstacles[i].y)) {
				food_count--;
				set_food();	
			}
		}
	}
	
}


/* Init snake */
void init_snake(void)
{
	unsigned int i;
	
	direction = LEFT;
	food_count = 0;
	score = 0;
	snake_size = SNAKE_START_LENGTH;
	
	set_food();

	for (i = 0; i < snake_size; i++) {
		snake[i].y = 3;
		snake[i].x = 6 + i;
	}
	
}


/* Init obstacles for more dificult levels */
void init_obstacles(void)
{
	unsigned int i;
	
	for (i = 0; i < NUM_OBSTACLES; i++) {
		obstacles[i].x = rand() % (WIDTH - 1) + 1;
		obstacles[i].y = rand() % (HEIGHT - 1) + 1;
	}
}


/* Check food or bonus or poison collisions */
void check_collect_goodies(void)
{	
	/* if food is collected */
	if (snake[0].x == food.x && snake[0].y == food.y) {
		int x = snake[snake_size - 1].x - snake[snake_size - 2].x;
		int y = snake[snake_size - 1].y - snake[snake_size - 2].y;
		snake[snake_size].x = snake[snake_size - 1].x + x;
		snake[snake_size].y = snake[snake_size - 1].y + y;
		
		score += food_value;
		if (food_value < 0)
			snake_size--;
		else
			snake_size++;

		/* if enough goodies collected, upgrade level */
		if ((food_count % LEVEL_UPGRADE == 0) &&
			(food_count > 0) &&
			(single_level == 0)) {
			if (level < 3)
				level++;
			init_snake();
			if (level > 1)
				init_obstacles();
		}
		else
			set_food();
	}
	
}


/* Take action for the end of the game */
void end_level(void)
{
	unsigned int i;
	char final_score[10];

	itoa(score, final_score, 10);
	lcd_clear();
	lcd_goto_xy(5, 3);
	for (i = 0; i < snake_size; i++) {
		snake[i].x = 0;
		snake[i].y = 0;
	}
	
	lcd_str("WASTED!");
	lcd_goto_xy(7, 4);
	lcd_str(final_score);

	init_snake();
	if (level > 1)
		init_obstacles();
	/* back to level 1 if you collide in arcade mode */
	if (single_level == 0)
		level = 1;
	_delay_ms(5000);
	
}


/* Check for end of the game */
void check_end_game(void)
{
	unsigned int i, j;
	
	for (i = 1; i < snake_size; i++) {
	
		/* collision with snake */
		if (snake[0].x == snake[i].x && snake[0].y == snake[i].y) {
			end_level();
			break;
		}
	}
		
	/* collision with obstacles for level 2 */
	if ((level == 2) || (level == 3)) {
		for (j = 0; j < NUM_OBSTACLES; j++) {
			if ((obstacles[j].x == snake[0].x) &&
				(obstacles[j].y == snake[0].y)) {
					end_level();
					break;
				}
		}
		if ((score < 0) && (food_count > 3))
			end_level();
	}

}


/* Get movement directions */
void get_direction(void)
{
    /* DOWN */
    if (((PINB & (1 << PB1)) != 0) && direction != UP)
        direction = DOWN;
	
    /* RIGHT */
    if (((PINB & (1 << PB3)) != 0) && direction != LEFT)
        direction = RIGHT;
		
    /* LEFT */
    if (((PINB & (1 << PB0)) != 0) && direction != RIGHT)
        direction = LEFT;
    
    /* UP */
    if (((PINB & (1 << PB2)) != 0) && direction != DOWN)
        direction = UP;
		
}


/* Move snake based on set direction */
void move_snake(void)
{
	unsigned int i;

	/* each piece of the snake gets the position of */
	/* the one in front of it; only head can change direction */
	for (i = snake_size - 1; i > 0; i--) {
		snake[i].x = snake[i - 1].x;
		snake[i].y = snake[i - 1].y;
	}

	if (direction == UP)
		snake[0].y = (snake[0].y - 1) ? (snake[0].y - 1) : HEIGHT;
	else if (direction == DOWN)
		snake[0].y = snake[0].y % HEIGHT + 1;
	else if (direction == LEFT)
		snake[0].x = (snake[0].x - 1 ) ? (snake[0].x - 1) : WIDTH;
	else if (direction == RIGHT)
		snake[0].x = snake[0].x % WIDTH + 1;

}


/* Draw game elements (snake, goodies, obstacles) */
void draw_game(void)
{
	unsigned int i;

	/* draw food or bonus or poision */
	lcd_goto_xy(food.x, food.y);
	if (food_value == FOOD_VALUE)
		lcd_chr('F');
	else if (food_value == BONUS_VALUE)
		lcd_chr('O');
	else
		lcd_chr('X');

	/* draw snake using '*' symbol */
	for (i = 0; i < snake_size; i++) {
		lcd_goto_xy(snake[i].x, snake[i].y);
		lcd_chr('*');
	}
	
	/* draw obstacles for levels 2 and 3 */
	if ((level == 2) || (level == 3)) {
		for (i = 0; i < NUM_OBSTACLES; i++) {
			lcd_goto_xy(obstacles[i].x, obstacles[i].y);
			lcd_chr('B');
		}
	}
	
}
 
 
 /* Selection menu at the start of the game */
void select_level(void)
{
    /* UP - level 1 */
    if (((PINB & (1 << PB2)) != 0) && level == 0) {
        level = 1;
		single_level = 1;
    }

	/* LEFT - level 2 */
    if (((PINB & (1 << PB0)) != 0) && level == 0) {
        level = 2;
		single_level = 1;
    }
	
    /* RIGHT - level 3 */
    if (((PINB & (1 << PB3)) != 0) && level == 0) {
        level = 3;
		single_level = 1;
	}

     /* DOWN - play all levels starting from level 1 */
    if (((PINB & (1 << PB1)) != 0) && level == 0)
        level = 1;
		
}


/* Init game parameters and start menu */
void init_game(void)
{
	/* Setup LCD */
	lcd_init();
	lcd_contrast(0x40);

	lcd_goto_xy(1, 2);
	lcd_str("Just Snake IT!");
	_delay_ms(2000);
	lcd_goto_xy(1, 4);
	lcd_str("Diaconu Cristian 333CA");
	_delay_ms(2000);
	
	lcd_clear();
	
	lcd_goto_xy(2, 3);
	lcd_str("Select desired level:");
	_delay_ms(3000);
	
	lcd_clear();

	lcd_goto_xy(1, 2);
	lcd_str("LVL 1 - UP");
	lcd_goto_xy(1, 3);
	lcd_str("LVL 2 - LEFT");
	lcd_goto_xy(1, 4);
	lcd_str("LVL 3 - RIGHT");
	lcd_goto_xy(1, 5);
	lcd_str("ARCADE - DOWN");
	
	_delay_ms(2000);

	enable_buttons();
	init_snake();

	level = 0;
	single_level = 0;

	/* wait for level selection */
	while (1) {
		select_level();
		_delay_ms(100);
		if (level > 0)
			break;
	}
	
	if (level > 1)
		init_obstacles();

	lcd_clear();
	
}


/* Application entry point */
int main(void)
{	
	
	init_game();

	while(1) {
		draw_game();
		get_direction();
		_delay_ms(150);
		lcd_clear();
		move_snake();
		check_collect_goodies();
		check_end_game();
	}
	
}
